﻿using Rhino;
using Rhino.Commands;
using Rhino.Geometry;
using Rhino.Input;
using Rhino.Input.Custom;
using System.Drawing;
using Rhino.Display;
using Rhino.Geometry.Intersect;
using System.Collections.Generic;

namespace PatrickCurve
{
    public class PatrickCurveCommand : Command
    {
        private class PatrickCurveConduit : DisplayConduit
        {
            private List<Line> _lines;
            public PatrickCurveConduit()
            {
                _lines = new List<Line>();
            }
            public void AddLine(Line line)
            {
                _lines.Add(line);
            }
            protected override void CalculateBoundingBox(CalculateBoundingBoxEventArgs e)
            {
                var bbox = new BoundingBox();
                foreach (var line in _lines)
                    bbox.Union(line.BoundingBox);
                e.IncludeBoundingBox(bbox);
            }
            protected override void DrawOverlay(DrawEventArgs e)
            {
                e.Display.DrawLines(_lines, Color.DarkRed);
            }
        }
        public PatrickCurveCommand()
        {
            Instance = this;
        } public static PatrickCurveCommand Instance
        {
            get; private set;
        }
        public override string EnglishName => "PatrickCurve";
        protected override Result RunCommand(RhinoDoc doc, RunMode mode)
        {
            var conduit = new PatrickCurveConduit { Enabled = true };
            var tol = doc.ModelAbsoluteTolerance;
            var pts = new Point3d[6];
            using (var getPointAction = new GetPoint())
            {
                getPointAction.SetCommandPrompt("Please select P0");
                if (getPointAction.Get() != GetResult.Point)
                {
                    conduit.Enabled = false;
                    doc.Views.Redraw();
                    RhinoApp.WriteLine("No point was selected.");
                    return getPointAction.CommandResult();
                }
                pts[0] = getPointAction.Point();
            }
            using (var getPointAction = new GetPoint())
            {
                getPointAction.SetCommandPrompt("Please select P5");
                getPointAction.SetBasePoint(pts[0], true);
                getPointAction.DynamicDraw +=
                  (sender, e) => e.Display.DrawLine(pts[0], e.CurrentPoint, Color.Black);
                if (getPointAction.Get() != GetResult.Point)
                {
                    conduit.Enabled = false;
                    doc.Views.Redraw();
                    RhinoApp.WriteLine("No point was selected.");
                    return getPointAction.CommandResult();
                }
                pts[5] = getPointAction.Point();
            }
            var baseLine = new Line(pts[0], pts[5]);
            conduit.AddLine(baseLine);
            doc.Views.Redraw();
            baseLine.ToNurbsCurve().PerpendicularFrameAt(0.5, out var perpFrame);
            var perpLine = new Line(perpFrame.Origin, perpFrame.XAxis);
            Point3d pt;
            using (var getPointAction = new GetPoint())
            {
                getPointAction.SetCommandPrompt("Please select P");
                getPointAction.SetBasePoint(perpLine.From, true);
                getPointAction.Constrain(perpLine);
                getPointAction.DynamicDraw +=
                    (sender, e) => e.Display.DrawLine(pts[0], pts[5], Color.Black);
                if (getPointAction.Get() != GetResult.Point)
                {
                    conduit.Enabled = false;
                    doc.Views.Redraw();
                    RhinoApp.WriteLine("No point was selected.");
                    return getPointAction.CommandResult();
                }
                pt = getPointAction.Point();
            }
            conduit.AddLine(new Line(perpLine.From, pt));
            doc.Views.Redraw();
            using (var getPointAction = new GetPoint())
            {
                getPointAction.SetCommandPrompt("Please select P4");
                getPointAction.SetBasePoint(pts[5], true);
                getPointAction.DynamicDraw +=
                    (sender, e) => e.Display.DrawLine(pts[5], e.CurrentPoint, Color.Black);
                if (getPointAction.Get() != GetResult.Point)
                {
                    conduit.Enabled = false;
                    doc.Views.Redraw();
                    RhinoApp.WriteLine("No point was selected.");
                    return getPointAction.CommandResult();
                }
                pts[4] = getPointAction.Point();
            }
            conduit.AddLine(new Line(pts[5], pts[4]));
            doc.Views.Redraw();
            using (var getPointAction = new GetPoint())
            {
                getPointAction.SetCommandPrompt("Please select P1");
                getPointAction.SetBasePoint(pts[0], true);
                getPointAction.DynamicDraw +=
                    (sender, e) => e.Display.DrawLine(pts[0], e.CurrentPoint, Color.Black);
                if (getPointAction.Get() != GetResult.Point)
                {
                    conduit.Enabled = false;
                    doc.Views.Redraw();
                    RhinoApp.WriteLine("No point was selected.");
                    return getPointAction.CommandResult();
                }
                pts[1] = getPointAction.Point();
            }
            conduit.AddLine(new Line(pts[0], pts[1]));
            doc.Views.Redraw();
            
            using (var getPointAction = new GetPoint())
            {
                var topLine = new Line(pt, perpFrame.ZAxis);
                getPointAction.SetCommandPrompt("Please select P2");
                getPointAction.SetBasePoint(pt, true);
                getPointAction.Constrain(topLine);
                getPointAction.DynamicDraw += (sender, e) =>
                {
                    pts[2] = e.CurrentPoint;
                    var curve = ComputeCurve(pts, pt, tol * 2);
                    e.Display.DrawPolyline(new Polyline(new List<Point3d>()
                    {
                        pts[1], curve.Points[2].Location, curve.Points[3].Location, pts[4]
                    }), Color.DarkRed);
                    e.Display.DrawCurve(curve, Color.Red, 2);
                    
                };

                if (getPointAction.Get() != GetResult.Point)
                {
                    conduit.Enabled = false;
                    doc.Views.Redraw();
                    RhinoApp.WriteLine("No point was selected.");
                    return getPointAction.CommandResult();
                }
                pts[2] = getPointAction.Point();
            } 
            doc.Objects.AddCurve(ComputeCurve(pts, pt, tol));
            conduit.Enabled = false;
            doc.Views.Redraw();
            return Result.Success;
        }
        private NurbsCurve ComputeCurve(Point3d[] pts,Point3d p, double tol)
        {
            var baseLine = new Line(pts[0], pts[5]).ToNurbsCurve();
            baseLine.PerpendicularFrameAt(0.5, out var perpFrame);
            var perpLine = new Line(perpFrame.Origin, p);
            baseLine.FrameAt(0.5, out var horiFrame);
            var tempPt = pts[2];
            tempPt.Transform(Transform.Mirror(perpFrame));
            pts[3] = tempPt;
            var c = NurbsCurve.Create(false, 5, pts);
            for (var i = 0; i < 300; i++)
            {
                c.GetBoundingBox(horiFrame, out var bb);
                var edge = new Rectangle3d(bb.Plane, bb.X, bb.Y).ToPolyline().SegmentAt(0);
                var tp = edge.ClosestPoint(p, true);
                var v = new Vector3d(p - tp);
                pts[2] += v;
                pts[3] += v;
                tp = Intersection.CurveLine(c, edge, tol, tol)[0].PointA;
                var tpPerpToPerpLine = perpLine.ClosestPoint(tp , true); 
                pts[3] += new Vector3d(tpPerpToPerpLine - tp);
                c = NurbsCurve.Create(false, 5, pts);
                if (p.DistanceTo(tp) < tol) break;
            }
            return c;
        }
    }
    
}